#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdarg.h>
#include <sys/mman.h>
#include <sys/kd.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/vt.h>
#include <sys/wait.h>
#include <errno.h>
#include <ctype.h>
#include "radeon_dump.h"

#define PCI_VENDOR_ID_ATI 0x1002

static pci_device_id radeonov_pci_table[] = {
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_RADEON_QD, RADEON_QD},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_RADEON_QE, RADEON_QE},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_RADEON_QF, RADEON_QF},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_RADEON_QG, RADEON_QG},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_RV100_QY,  RV100_QY},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_RV100_QZ,  RV100_QZ},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_RV100_LY,  RV100_LY},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_RV100_LZ,  RV100_LZ},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_RS100_4136,RS100_4136},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_RS100_4336,RS100_4336},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_RS200_4137,RS200_4137},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_RS200_4337,RS200_4337},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_RS250_4237,RS250_4237},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_RS250_4437,RS250_4437},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_RV200_QW,  RV200_QW},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_RV200_QX,  RV200_QX},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_RV200_LW,  RV200_LW},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_RV200_LX,  RV200_LX},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_R200_QH,   R200_QH},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_R200_QI,   R200_QI},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_R200_QJ,   R200_QJ},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_R200_QK,   R200_QK},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_R200_QL,   R200_QL},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_R200_QM,   R200_QM},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_R200_QN,   R200_QN},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_R200_QO,   R200_QO},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_R200_BB,   R200_BB},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_R200_BC,   R200_BC},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_RV250_Id,  RV250_Id},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_RV250_Ie,  RV250_Ie},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_RV250_If,  RV250_If},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_RV250_Ig,  RV250_Ig},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_RV250_Ld,  RV250_Ld},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_RV250_Le,  RV250_Le},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_RV250_Lf,  RV250_Lf},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_RV250_Lg,  RV250_Lg},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_RS300_5834,RS300_5834},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_RS300_5835,RS300_5835},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_RV280_5960,RV280_5960},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_RV280_5961,RV280_5961},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_RV280_5962,RV280_5962},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_RV280_5963,RV280_5963},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_RV280_5964,RV280_5964},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_RV280_5C61,RV280_5C61},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_RV280_5C62,RV280_5C62},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_R300_AD,   R300_AD},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_R300_AE,   R300_AE},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_R300_AF,   R300_AF},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_R300_AG,   R300_AG},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_R300_ND,   R300_ND},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_R300_NE,   R300_NE},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_R300_NF,   R300_NF},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_R300_NG,   R300_NG},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_R350_AH,   R350_AH},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_R350_AI,   R350_AI},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_R350_AJ,   R350_AJ},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_R350_AK,   R350_AK},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_R350_NH,   R350_NH},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_R350_NI,   R350_NI},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_R350_NJ,   R350_NJ},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_R350_NK,   R350_NK},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_RV350_AP,  RV350_AP},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_RV350_AQ,  RV350_AQ},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_RV350_AR,  RV350_AR},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_RV350_AS,  RV350_AS},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_RV350_AT,  RV350_AT},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_RV350_AV,  RV350_AV},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_RV350_NP,  RV350_NP},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_RV350_NQ,  RV350_NQ},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_RV350_NR,  RV350_NR},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_RV350_NS,  RV350_NS},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_RV350_NT,  RV350_NT},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_RV350_NV,  RV350_NV},
    { PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_R420_JK,  R420_JK},
    { 0, 0, 0}
};

static char* ChipName[LAST_CHIP+1] = {
    "R100 QD (7200)",
    "R100 QE (7200)",
    "R100 QF (7200)",
    "R100 QG (7200)",
    "RV100 QY (7000)",
    "RV100 QZ (7000)",
    "RV100 LY (M6)",
    "RV100 LZ (M6)",
    "RS100 4136 (IGP320(A3))",
    "RS100 4336 (IGP320(U1))",
    "RS200 4137 (IGP340(A4))",
    "RS200 4337 (IGP340(U2))",
    "RS250 4237 (IGP 7000 (A4+))",
    "RS250 4437 (Mobility IGP 7000",
    "RV200 QW (7500)",
    "RV200 QX (7500)",
    "RV200 LW (M7)",
    "RV200 LX (FireGL 7800 M7)",
    "R200 QH (FireGL 8700/8800)",
    "R200 QI (8500)",
    "R200 QJ (8500)",
    "R200 QK (8500)",
    "R200 QL (8500)",
    "R200 QM (8500)",
    "R200 QN (8500)",
    "R200 QO (8500)",
    "R200 BB (8500 AIW)",
    "R200 BC (8500 AIW)",
    "RV250 Id (9000)",
    "RV250 Ie (9000)",
    "RV250 If (9000PRO)",
    "RV250 Ig (9000)",
    "RV250 Ld (FireGL Mobility 9000 (M9))",
    "RV250 Le (Mobility 9000 (M9))",
    "RV250 Lf (Mobility 9000 (M9))",
    "RV250 Lg (Mobility 9000 (M9))",
    "RS300 5834 (IGP 9100 (A5))",
    "RS300 5835 (IGP Mobility 9100 (U3))",
    "RV280 5960 (9200)",
    "RV280 5961 (9200)",
    "RV280 5962 (9200)",
    "RV280 5963 (9200)",
    "RV280 5964 (9200SE)",
    "RV280 5C61 (M9+)",
    "RV280 5C63 (M9+)",
    "R300 AD (9500)",
    "R300 AE (9500)",
    "R300 AF (9600TX)",
    "R300 AG (FireGL Z1)",
    "R300 ND (9700PRO)",
    "R300 NE (9700/9500PRO)",
    "R300 NF (9700)",
    "R300 NG (FireGL X1)",
    "R350 AH (9800SE)",
    "R350 AI (9800)",
    "R350 AJ (9800)",
    "R350 AK (FireGL X2)",
    "R350 NH (9800PRO)",
    "R350 NI (9800)",
    "R350 NJ (9800XT)",
    "R350 NK (FireGL X2)",
    "RV350 AP (9600)",
    "RV350 AQ (9600SE)",
    "RV350 AR (9600XT)",
    "RV350 AS (9600)",
    "RV350 AT (FireGL T2)",
    "RV350 AV (FireGL RV360)",
    "RV350 NP (Mobility 9600 (M10))",
    "RV350 NQ (Mobility 9600 (M10))",
    "RV350 NR (Mobility 9600 (M11))",
    "RV350 NS (Mobility 9600 (M10))",
    "RV350 NT (FireGL Mobility T2 (M10))",
    "RV350 NV (FireGL Mobility T2 (M11))",
    "R420 JK (X800 XT VIVO)",
    "UNKNOWN Chip"
};

static int proc_pci_read_config(unsigned char bus, unsigned char device,
                                unsigned char fn, unsigned long *buf, int size)
{
    int f,i;
    char filename[256];

    sprintf(filename,"/proc/bus/pci/%02i/%02x.%i",bus,device,fn);
    f=open(filename,O_RDONLY);
    if(read(f,buf,4*size)<1)for(i=0;i<63;i++)buf[i]=-1;
    close(f);
    return 0;
};

static int pci_find_radeon(unsigned long *conf)
{ 
    unsigned long buf[64];
    int bus, device=0, i;
    unsigned long id;

    for(bus=0; (bus<16); bus++) {
        for(device=0; device<32; device++) {
            proc_pci_read_config(bus, device, 0, buf, 3);
            if(((buf[0]&0xffff)==PCI_VENDOR_ID_ATI)&&
               (((buf[2]>>16)&0xffff)==0x0300)) {
                id = buf[0]>>16;
                i = 0;
                while (radeonov_pci_table[i].VendorID) {
                    if (radeonov_pci_table[i].DeviceID == id) break;
                    i++;
                }
                if(!radeonov_pci_table[i].VendorID) continue;
                proc_pci_read_config(bus, device, 0, buf, 64);
                memcpy(conf,buf,256);
                return i;
            }
        };
    }
    return -1;
}

static int radeon_init(void)
{
    int idx;

    rinfo = malloc(sizeof(radeonov_info));
    if (!rinfo) {
        fprintf (stderr, "Cannot allocate rinof\n");
        return 0;
    }
    memset (rinfo, 0, sizeof (radeonov_info));

    if ((rinfo->fd = open("/dev/mem", O_RDWR)) < 0) {
        fprintf(stderr, "Cannot open /dev/mem\n" \
               "You need to have root priviledge to use this library\n");
        return 0;
    }
    
    if((idx = pci_find_radeon(rinfo->pci_buf)) < 0) {
        fprintf(stderr, "Unknown card\n");
    }

    rinfo->Chipset = (rinfo->pci_buf[0]>>16) & 0xffff;

    rinfo->mmio_base_phys = rinfo->pci_buf[6] & 0xffffff00;

    rinfo->mmio_base = (u32) mmap(0, 20*1024, PROT_READ | PROT_WRITE, 
                        MAP_SHARED, rinfo->fd, rinfo->mmio_base_phys);
    if (!rinfo->mmio_base) {
        fprintf (stderr, "radeonov: cannot map MMIO\n");
        return 0;
    }

    rinfo->bios_seg  = (char*) mmap(0, 64*1024, PROT_READ | PROT_WRITE, 
                        MAP_SHARED, rinfo->fd, 0xc0000);
    if (!rinfo->bios_seg) {
        fprintf (stderr, "radeonov: cannot map BIOS\n");
        munmap((char*)rinfo->mmio_base, 16*1024);
        free(rinfo);
        close(rinfo->fd);
        return 0;
    }

    printf ("radeon_dump: ATI %s\n", ChipName[idx]);

    return 1;
}

static int radeon_close(void)
{
    munmap((char*)rinfo->mmio_base, 16*1024);
    munmap(rinfo->bios_seg, 64*1024);
    free(rinfo);
    close(rinfo->fd);
    return 1;
}

static void dump_bios()
{
    int i, j;
    u8 tmp;
    printf("BIOS Image start: -------------------\n");
    for (i=0; i<1024; i++) {
	for (j=0; j<16; j++) {
	    printf("%02x ", *(u8*)(rinfo->bios_seg+i*16+j));
	}
	printf("  ");
	for (j=0; j<16; j++) {
	    tmp = *(u8*)(rinfo->bios_seg+i*16+j);
	    if (tmp<32 || tmp>127) tmp='.';
	    printf("%c", tmp);
	}
	printf("\n");
    }
    printf("BIOS Image End: -------------------\n");
}

static void dump_pci()
{
    int i, j;
    u32 tmp;
    printf("PCI Configuration Start: -------------------\n");
    for (i=0; i<16; i++) {
	printf("%02x: ", i*16);
	for (j=0; j<4; j++) {
	    tmp = rinfo->pci_buf[i*4+j];
	    printf("%02lx %02lx %02lx %02lx ", 
		   tmp & 0xff, (tmp >> 8) & 0xff,
		   (tmp >> 16) & 0xff, (tmp >> 24) & 0xff);
	}
	printf("\n");
    }
    printf("PCI Configuration End: -------------------\n");
}

static void dump_reg()
{
    int i, j;
    printf("------------ Dumping MMIO registers -------------\n");
    
   for (i=0x0; i<0x1FFF; i += 4){
	printf("0x%08x ", i);
        printf("0x%08lx", INREG(i));
	printf("\n");
    }
/*   for (i=0x0/16; i<(0x4FF0/16); i++) {
	printf("%04x: ", i*16);
	for (j=0; j<16; j=j+4) {
	    printf("%08lx ", INREG(i*16+j));
	}
	printf("\n");
    }
*/

//    printf("%08lx ", INREG(0x4FF0));
/*    i = 0x4310;
    printf("%04x: ", i);
    for (j=0; j<8; j++) {
        printf("%08lx ", INREG(i+j*4));
    }
    printf("\n");
*/
/*    printf("%04x: ", i);
    for (j=0; j<16; j=j+4) {
        printf("%08x ", 1.0f);
    }*/
/*    printf("________ ");
    printf("%08lx ", INREG(0x4FF0+16));*/
//    printf("\nMMIO Registers End: -------------------\n");
}

static void dump_pll()
{
    int i, j;
    printf("------------ Dumping PLL registers -------------\n");
    for (i=0; i<0x24/4; i++) {
	for (j=0; j<4; j++) {
	    printf(" 0x%08lx", INPLL(i*4+j));
	}
	printf("\n");
    }
//    printf("PLL Registers Start: -------------------\n");
}

int main(int argc, char** argv)
{
    if(!radeon_init()) return 0;

    //dump_bios();
    //dump_pci();
    dump_pll();
    dump_reg();

    radeon_close();
    return 0;
}
